/*
 * Author: Kipp Cannon <kipp@gravity.phys.uwm.edu>
 * Revision: $Id: Sequence.m4,v 1.1 2005/01/28 23:38:06 kipp Exp $
 */

#include <string.h>
#include <lal/Date.h>
#include <lal/LALDatatypes.h>
#include <lal/LALStdlib.h>
#include <lal/LALStatusMacros.h>
#include <lal/LALErrno.h>
#include <lal/LALError.h>
#include <lal/Sequence.h>

NRCSID(SEQUENCEC, "$Id: Sequence.m4,v 1.1 2005/01/28 23:38:06 kipp Exp $");

/*
 * Shift the bytes in the buffer buff, whose length is length, count bytes to
 * higher addresses.  If the magnitude of count is greater than or equal to
 * that of length, then nothing is done.
 */

static void memshift(void *buff, size_t length, int count)
{
	if(count >= 0) {
		if(length > (size_t) count)
			memmove((char *) buff + count, buff, length - count);
	} else {
		if(length > (size_t) -count)
			memmove(buff, (char *) buff - count, length + count);
	}
}


/*
 * Begin library functions...
 */



void XLALDestroyREAL4Sequence (
	REAL4Sequence *sequence
)
{
	if(sequence)
		LALFree(sequence->data);
	LALFree(sequence);
}


void LALDestroyREAL4Sequence (
	LALStatus *status,
	REAL4Sequence *sequence
)
{
	INITSTATUS(status, "LALDestroyREAL4Sequence", SEQUENCEC);
	XLALDestroyREAL4Sequence (sequence);
	RETURN(status);
}


REAL4Sequence *XLALCreateREAL4Sequence (
	size_t length
)
{
	REAL4Sequence *new;
	REAL4 *data;

	new = LALMalloc(sizeof(*new));
	data = LALMalloc(length * sizeof(*data));
	if(!new || !data) {
		LALFree(new);
		LALFree(data);
		return(NULL);
	}

	new->data = data;
	new->length = length;

	return(new);
}


void LALCreateREAL4Sequence (
	LALStatus *status,
	REAL4Sequence **output,
	size_t length
)
{
	INITSTATUS(status, "LALCreateREAL4Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateREAL4Sequence (length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


REAL4Sequence *XLALCutREAL4Sequence (
	REAL4Sequence *sequence,
	size_t first,
	size_t length
)
{
	REAL4Sequence *new = NULL;

	if(sequence && sequence->data) {
		new = XLALCreateREAL4Sequence (length);
		if(new)
			memcpy(new->data, sequence->data + first, length * sizeof(*new->data));
	}

	return(new);
}


void LALCutREAL4Sequence (
	LALStatus *status,
	REAL4Sequence **output,
	REAL4Sequence *input,
	size_t first,
	size_t length
)
{
	INITSTATUS(status, "LALCutREAL4Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= input->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	*output = XLALCutREAL4Sequence (input, first, length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


REAL4Sequence *XLALCopyREAL4Sequence (
	REAL4Sequence *sequence
)
{
	return(XLALCutREAL4Sequence (sequence, 0, sequence->length));
}


void LALCopyREAL4Sequence (
	LALStatus *status,
	REAL4Sequence **output,
	REAL4Sequence *input
)
{
	INITSTATUS(status, "LALCopyREAL4Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCopyREAL4Sequence (input);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


static void ShiftREAL4Sequence (
	REAL4Sequence *sequence,
	int count
)
{
	memshift(sequence->data, sequence->length * sizeof(*sequence->data), count * (int) sizeof(*sequence->data));
}


void XLALShiftREAL4Sequence (
	REAL4Sequence *sequence,
	int count
)
{
	if(!sequence || !sequence->data || !count)
		return;

	ShiftREAL4Sequence (sequence, count);
	if((size_t) labs(count) >= sequence->length)
		memset(sequence->data, 0, sequence->length * sizeof(*sequence->data));
	else if(count > 0)
		memset(sequence->data, 0, count * sizeof(*sequence->data));
	else
		memset(sequence->data + sequence->length + count, 0, -count * sizeof(*sequence->data));
}


void LALShiftREAL4Sequence (
	LALStatus *status,
	REAL4Sequence *sequence,
	int count
)
{
	INITSTATUS(status, "LALShiftREAL4Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	XLALShiftREAL4Sequence (sequence, count);
	RETURN(status);
}


/* FIXME: this function does not take care to move the least number of bytes
 * possible.  A performance gain would be realized by being more careful. */
size_t XLALResizeREAL4Sequence (
	REAL4Sequence *sequence,
	int first,
	size_t length
)
{
	if(!sequence || !sequence->data)
		return(0);

	if(length > sequence->length) {
		/* need to increase memory */
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));

		if(sequence->data)
			ShiftREAL4Sequence (sequence, -first);
		else
			length = 0;
	} else {
		/* do not need to increase memory */
		ShiftREAL4Sequence (sequence, -first);
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));
		if(!sequence->data)
			length = 0;
	}

	return(sequence->length = length);
}


void LALResizeREAL4Sequence (
	LALStatus *status,
	REAL4Sequence *sequence,
	int first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALResizeREAL4Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	new_length = XLALResizeREAL4Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}


size_t XLALShrinkREAL4Sequence (
	REAL4Sequence *sequence,
	size_t first,
	size_t length
)
{
	return(XLALResizeREAL4Sequence (sequence, first, length));
}


void LALShrinkREAL4Sequence (
	LALStatus *status,
	REAL4Sequence *sequence,
	size_t first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALShrinkREAL4Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= sequence->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	new_length = XLALShrinkREAL4Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}




void XLALDestroyREAL8Sequence (
	REAL8Sequence *sequence
)
{
	if(sequence)
		LALFree(sequence->data);
	LALFree(sequence);
}


void LALDestroyREAL8Sequence (
	LALStatus *status,
	REAL8Sequence *sequence
)
{
	INITSTATUS(status, "LALDestroyREAL8Sequence", SEQUENCEC);
	XLALDestroyREAL8Sequence (sequence);
	RETURN(status);
}


REAL8Sequence *XLALCreateREAL8Sequence (
	size_t length
)
{
	REAL8Sequence *new;
	REAL8 *data;

	new = LALMalloc(sizeof(*new));
	data = LALMalloc(length * sizeof(*data));
	if(!new || !data) {
		LALFree(new);
		LALFree(data);
		return(NULL);
	}

	new->data = data;
	new->length = length;

	return(new);
}


void LALCreateREAL8Sequence (
	LALStatus *status,
	REAL8Sequence **output,
	size_t length
)
{
	INITSTATUS(status, "LALCreateREAL8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateREAL8Sequence (length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


REAL8Sequence *XLALCutREAL8Sequence (
	REAL8Sequence *sequence,
	size_t first,
	size_t length
)
{
	REAL8Sequence *new = NULL;

	if(sequence && sequence->data) {
		new = XLALCreateREAL8Sequence (length);
		if(new)
			memcpy(new->data, sequence->data + first, length * sizeof(*new->data));
	}

	return(new);
}


void LALCutREAL8Sequence (
	LALStatus *status,
	REAL8Sequence **output,
	REAL8Sequence *input,
	size_t first,
	size_t length
)
{
	INITSTATUS(status, "LALCutREAL8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= input->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	*output = XLALCutREAL8Sequence (input, first, length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


REAL8Sequence *XLALCopyREAL8Sequence (
	REAL8Sequence *sequence
)
{
	return(XLALCutREAL8Sequence (sequence, 0, sequence->length));
}


void LALCopyREAL8Sequence (
	LALStatus *status,
	REAL8Sequence **output,
	REAL8Sequence *input
)
{
	INITSTATUS(status, "LALCopyREAL8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCopyREAL8Sequence (input);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


static void ShiftREAL8Sequence (
	REAL8Sequence *sequence,
	int count
)
{
	memshift(sequence->data, sequence->length * sizeof(*sequence->data), count * (int) sizeof(*sequence->data));
}


void XLALShiftREAL8Sequence (
	REAL8Sequence *sequence,
	int count
)
{
	if(!sequence || !sequence->data || !count)
		return;

	ShiftREAL8Sequence (sequence, count);
	if((size_t) labs(count) >= sequence->length)
		memset(sequence->data, 0, sequence->length * sizeof(*sequence->data));
	else if(count > 0)
		memset(sequence->data, 0, count * sizeof(*sequence->data));
	else
		memset(sequence->data + sequence->length + count, 0, -count * sizeof(*sequence->data));
}


void LALShiftREAL8Sequence (
	LALStatus *status,
	REAL8Sequence *sequence,
	int count
)
{
	INITSTATUS(status, "LALShiftREAL8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	XLALShiftREAL8Sequence (sequence, count);
	RETURN(status);
}


/* FIXME: this function does not take care to move the least number of bytes
 * possible.  A performance gain would be realized by being more careful. */
size_t XLALResizeREAL8Sequence (
	REAL8Sequence *sequence,
	int first,
	size_t length
)
{
	if(!sequence || !sequence->data)
		return(0);

	if(length > sequence->length) {
		/* need to increase memory */
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));

		if(sequence->data)
			ShiftREAL8Sequence (sequence, -first);
		else
			length = 0;
	} else {
		/* do not need to increase memory */
		ShiftREAL8Sequence (sequence, -first);
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));
		if(!sequence->data)
			length = 0;
	}

	return(sequence->length = length);
}


void LALResizeREAL8Sequence (
	LALStatus *status,
	REAL8Sequence *sequence,
	int first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALResizeREAL8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	new_length = XLALResizeREAL8Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}


size_t XLALShrinkREAL8Sequence (
	REAL8Sequence *sequence,
	size_t first,
	size_t length
)
{
	return(XLALResizeREAL8Sequence (sequence, first, length));
}


void LALShrinkREAL8Sequence (
	LALStatus *status,
	REAL8Sequence *sequence,
	size_t first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALShrinkREAL8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= sequence->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	new_length = XLALShrinkREAL8Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}




void XLALDestroyCOMPLEX8Sequence (
	COMPLEX8Sequence *sequence
)
{
	if(sequence)
		LALFree(sequence->data);
	LALFree(sequence);
}


void LALDestroyCOMPLEX8Sequence (
	LALStatus *status,
	COMPLEX8Sequence *sequence
)
{
	INITSTATUS(status, "LALDestroyCOMPLEX8Sequence", SEQUENCEC);
	XLALDestroyCOMPLEX8Sequence (sequence);
	RETURN(status);
}


COMPLEX8Sequence *XLALCreateCOMPLEX8Sequence (
	size_t length
)
{
	COMPLEX8Sequence *new;
	COMPLEX8 *data;

	new = LALMalloc(sizeof(*new));
	data = LALMalloc(length * sizeof(*data));
	if(!new || !data) {
		LALFree(new);
		LALFree(data);
		return(NULL);
	}

	new->data = data;
	new->length = length;

	return(new);
}


void LALCreateCOMPLEX8Sequence (
	LALStatus *status,
	COMPLEX8Sequence **output,
	size_t length
)
{
	INITSTATUS(status, "LALCreateCOMPLEX8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateCOMPLEX8Sequence (length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


COMPLEX8Sequence *XLALCutCOMPLEX8Sequence (
	COMPLEX8Sequence *sequence,
	size_t first,
	size_t length
)
{
	COMPLEX8Sequence *new = NULL;

	if(sequence && sequence->data) {
		new = XLALCreateCOMPLEX8Sequence (length);
		if(new)
			memcpy(new->data, sequence->data + first, length * sizeof(*new->data));
	}

	return(new);
}


void LALCutCOMPLEX8Sequence (
	LALStatus *status,
	COMPLEX8Sequence **output,
	COMPLEX8Sequence *input,
	size_t first,
	size_t length
)
{
	INITSTATUS(status, "LALCutCOMPLEX8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= input->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	*output = XLALCutCOMPLEX8Sequence (input, first, length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


COMPLEX8Sequence *XLALCopyCOMPLEX8Sequence (
	COMPLEX8Sequence *sequence
)
{
	return(XLALCutCOMPLEX8Sequence (sequence, 0, sequence->length));
}


void LALCopyCOMPLEX8Sequence (
	LALStatus *status,
	COMPLEX8Sequence **output,
	COMPLEX8Sequence *input
)
{
	INITSTATUS(status, "LALCopyCOMPLEX8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCopyCOMPLEX8Sequence (input);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


static void ShiftCOMPLEX8Sequence (
	COMPLEX8Sequence *sequence,
	int count
)
{
	memshift(sequence->data, sequence->length * sizeof(*sequence->data), count * (int) sizeof(*sequence->data));
}


void XLALShiftCOMPLEX8Sequence (
	COMPLEX8Sequence *sequence,
	int count
)
{
	if(!sequence || !sequence->data || !count)
		return;

	ShiftCOMPLEX8Sequence (sequence, count);
	if((size_t) labs(count) >= sequence->length)
		memset(sequence->data, 0, sequence->length * sizeof(*sequence->data));
	else if(count > 0)
		memset(sequence->data, 0, count * sizeof(*sequence->data));
	else
		memset(sequence->data + sequence->length + count, 0, -count * sizeof(*sequence->data));
}


void LALShiftCOMPLEX8Sequence (
	LALStatus *status,
	COMPLEX8Sequence *sequence,
	int count
)
{
	INITSTATUS(status, "LALShiftCOMPLEX8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	XLALShiftCOMPLEX8Sequence (sequence, count);
	RETURN(status);
}


/* FIXME: this function does not take care to move the least number of bytes
 * possible.  A performance gain would be realized by being more careful. */
size_t XLALResizeCOMPLEX8Sequence (
	COMPLEX8Sequence *sequence,
	int first,
	size_t length
)
{
	if(!sequence || !sequence->data)
		return(0);

	if(length > sequence->length) {
		/* need to increase memory */
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));

		if(sequence->data)
			ShiftCOMPLEX8Sequence (sequence, -first);
		else
			length = 0;
	} else {
		/* do not need to increase memory */
		ShiftCOMPLEX8Sequence (sequence, -first);
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));
		if(!sequence->data)
			length = 0;
	}

	return(sequence->length = length);
}


void LALResizeCOMPLEX8Sequence (
	LALStatus *status,
	COMPLEX8Sequence *sequence,
	int first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALResizeCOMPLEX8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	new_length = XLALResizeCOMPLEX8Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}


size_t XLALShrinkCOMPLEX8Sequence (
	COMPLEX8Sequence *sequence,
	size_t first,
	size_t length
)
{
	return(XLALResizeCOMPLEX8Sequence (sequence, first, length));
}


void LALShrinkCOMPLEX8Sequence (
	LALStatus *status,
	COMPLEX8Sequence *sequence,
	size_t first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALShrinkCOMPLEX8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= sequence->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	new_length = XLALShrinkCOMPLEX8Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}




void XLALDestroyCOMPLEX16Sequence (
	COMPLEX16Sequence *sequence
)
{
	if(sequence)
		LALFree(sequence->data);
	LALFree(sequence);
}


void LALDestroyCOMPLEX16Sequence (
	LALStatus *status,
	COMPLEX16Sequence *sequence
)
{
	INITSTATUS(status, "LALDestroyCOMPLEX16Sequence", SEQUENCEC);
	XLALDestroyCOMPLEX16Sequence (sequence);
	RETURN(status);
}


COMPLEX16Sequence *XLALCreateCOMPLEX16Sequence (
	size_t length
)
{
	COMPLEX16Sequence *new;
	COMPLEX16 *data;

	new = LALMalloc(sizeof(*new));
	data = LALMalloc(length * sizeof(*data));
	if(!new || !data) {
		LALFree(new);
		LALFree(data);
		return(NULL);
	}

	new->data = data;
	new->length = length;

	return(new);
}


void LALCreateCOMPLEX16Sequence (
	LALStatus *status,
	COMPLEX16Sequence **output,
	size_t length
)
{
	INITSTATUS(status, "LALCreateCOMPLEX16Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateCOMPLEX16Sequence (length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


COMPLEX16Sequence *XLALCutCOMPLEX16Sequence (
	COMPLEX16Sequence *sequence,
	size_t first,
	size_t length
)
{
	COMPLEX16Sequence *new = NULL;

	if(sequence && sequence->data) {
		new = XLALCreateCOMPLEX16Sequence (length);
		if(new)
			memcpy(new->data, sequence->data + first, length * sizeof(*new->data));
	}

	return(new);
}


void LALCutCOMPLEX16Sequence (
	LALStatus *status,
	COMPLEX16Sequence **output,
	COMPLEX16Sequence *input,
	size_t first,
	size_t length
)
{
	INITSTATUS(status, "LALCutCOMPLEX16Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= input->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	*output = XLALCutCOMPLEX16Sequence (input, first, length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


COMPLEX16Sequence *XLALCopyCOMPLEX16Sequence (
	COMPLEX16Sequence *sequence
)
{
	return(XLALCutCOMPLEX16Sequence (sequence, 0, sequence->length));
}


void LALCopyCOMPLEX16Sequence (
	LALStatus *status,
	COMPLEX16Sequence **output,
	COMPLEX16Sequence *input
)
{
	INITSTATUS(status, "LALCopyCOMPLEX16Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCopyCOMPLEX16Sequence (input);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


static void ShiftCOMPLEX16Sequence (
	COMPLEX16Sequence *sequence,
	int count
)
{
	memshift(sequence->data, sequence->length * sizeof(*sequence->data), count * (int) sizeof(*sequence->data));
}


void XLALShiftCOMPLEX16Sequence (
	COMPLEX16Sequence *sequence,
	int count
)
{
	if(!sequence || !sequence->data || !count)
		return;

	ShiftCOMPLEX16Sequence (sequence, count);
	if((size_t) labs(count) >= sequence->length)
		memset(sequence->data, 0, sequence->length * sizeof(*sequence->data));
	else if(count > 0)
		memset(sequence->data, 0, count * sizeof(*sequence->data));
	else
		memset(sequence->data + sequence->length + count, 0, -count * sizeof(*sequence->data));
}


void LALShiftCOMPLEX16Sequence (
	LALStatus *status,
	COMPLEX16Sequence *sequence,
	int count
)
{
	INITSTATUS(status, "LALShiftCOMPLEX16Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	XLALShiftCOMPLEX16Sequence (sequence, count);
	RETURN(status);
}


/* FIXME: this function does not take care to move the least number of bytes
 * possible.  A performance gain would be realized by being more careful. */
size_t XLALResizeCOMPLEX16Sequence (
	COMPLEX16Sequence *sequence,
	int first,
	size_t length
)
{
	if(!sequence || !sequence->data)
		return(0);

	if(length > sequence->length) {
		/* need to increase memory */
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));

		if(sequence->data)
			ShiftCOMPLEX16Sequence (sequence, -first);
		else
			length = 0;
	} else {
		/* do not need to increase memory */
		ShiftCOMPLEX16Sequence (sequence, -first);
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));
		if(!sequence->data)
			length = 0;
	}

	return(sequence->length = length);
}


void LALResizeCOMPLEX16Sequence (
	LALStatus *status,
	COMPLEX16Sequence *sequence,
	int first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALResizeCOMPLEX16Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	new_length = XLALResizeCOMPLEX16Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}


size_t XLALShrinkCOMPLEX16Sequence (
	COMPLEX16Sequence *sequence,
	size_t first,
	size_t length
)
{
	return(XLALResizeCOMPLEX16Sequence (sequence, first, length));
}


void LALShrinkCOMPLEX16Sequence (
	LALStatus *status,
	COMPLEX16Sequence *sequence,
	size_t first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALShrinkCOMPLEX16Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= sequence->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	new_length = XLALShrinkCOMPLEX16Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}




void XLALDestroyINT2Sequence (
	INT2Sequence *sequence
)
{
	if(sequence)
		LALFree(sequence->data);
	LALFree(sequence);
}


void LALDestroyINT2Sequence (
	LALStatus *status,
	INT2Sequence *sequence
)
{
	INITSTATUS(status, "LALDestroyINT2Sequence", SEQUENCEC);
	XLALDestroyINT2Sequence (sequence);
	RETURN(status);
}


INT2Sequence *XLALCreateINT2Sequence (
	size_t length
)
{
	INT2Sequence *new;
	INT2 *data;

	new = LALMalloc(sizeof(*new));
	data = LALMalloc(length * sizeof(*data));
	if(!new || !data) {
		LALFree(new);
		LALFree(data);
		return(NULL);
	}

	new->data = data;
	new->length = length;

	return(new);
}


void LALCreateINT2Sequence (
	LALStatus *status,
	INT2Sequence **output,
	size_t length
)
{
	INITSTATUS(status, "LALCreateINT2Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateINT2Sequence (length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


INT2Sequence *XLALCutINT2Sequence (
	INT2Sequence *sequence,
	size_t first,
	size_t length
)
{
	INT2Sequence *new = NULL;

	if(sequence && sequence->data) {
		new = XLALCreateINT2Sequence (length);
		if(new)
			memcpy(new->data, sequence->data + first, length * sizeof(*new->data));
	}

	return(new);
}


void LALCutINT2Sequence (
	LALStatus *status,
	INT2Sequence **output,
	INT2Sequence *input,
	size_t first,
	size_t length
)
{
	INITSTATUS(status, "LALCutINT2Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= input->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	*output = XLALCutINT2Sequence (input, first, length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


INT2Sequence *XLALCopyINT2Sequence (
	INT2Sequence *sequence
)
{
	return(XLALCutINT2Sequence (sequence, 0, sequence->length));
}


void LALCopyINT2Sequence (
	LALStatus *status,
	INT2Sequence **output,
	INT2Sequence *input
)
{
	INITSTATUS(status, "LALCopyINT2Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCopyINT2Sequence (input);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


static void ShiftINT2Sequence (
	INT2Sequence *sequence,
	int count
)
{
	memshift(sequence->data, sequence->length * sizeof(*sequence->data), count * (int) sizeof(*sequence->data));
}


void XLALShiftINT2Sequence (
	INT2Sequence *sequence,
	int count
)
{
	if(!sequence || !sequence->data || !count)
		return;

	ShiftINT2Sequence (sequence, count);
	if((size_t) labs(count) >= sequence->length)
		memset(sequence->data, 0, sequence->length * sizeof(*sequence->data));
	else if(count > 0)
		memset(sequence->data, 0, count * sizeof(*sequence->data));
	else
		memset(sequence->data + sequence->length + count, 0, -count * sizeof(*sequence->data));
}


void LALShiftINT2Sequence (
	LALStatus *status,
	INT2Sequence *sequence,
	int count
)
{
	INITSTATUS(status, "LALShiftINT2Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	XLALShiftINT2Sequence (sequence, count);
	RETURN(status);
}


/* FIXME: this function does not take care to move the least number of bytes
 * possible.  A performance gain would be realized by being more careful. */
size_t XLALResizeINT2Sequence (
	INT2Sequence *sequence,
	int first,
	size_t length
)
{
	if(!sequence || !sequence->data)
		return(0);

	if(length > sequence->length) {
		/* need to increase memory */
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));

		if(sequence->data)
			ShiftINT2Sequence (sequence, -first);
		else
			length = 0;
	} else {
		/* do not need to increase memory */
		ShiftINT2Sequence (sequence, -first);
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));
		if(!sequence->data)
			length = 0;
	}

	return(sequence->length = length);
}


void LALResizeINT2Sequence (
	LALStatus *status,
	INT2Sequence *sequence,
	int first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALResizeINT2Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	new_length = XLALResizeINT2Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}


size_t XLALShrinkINT2Sequence (
	INT2Sequence *sequence,
	size_t first,
	size_t length
)
{
	return(XLALResizeINT2Sequence (sequence, first, length));
}


void LALShrinkINT2Sequence (
	LALStatus *status,
	INT2Sequence *sequence,
	size_t first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALShrinkINT2Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= sequence->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	new_length = XLALShrinkINT2Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}




void XLALDestroyUINT2Sequence (
	UINT2Sequence *sequence
)
{
	if(sequence)
		LALFree(sequence->data);
	LALFree(sequence);
}


void LALDestroyUINT2Sequence (
	LALStatus *status,
	UINT2Sequence *sequence
)
{
	INITSTATUS(status, "LALDestroyUINT2Sequence", SEQUENCEC);
	XLALDestroyUINT2Sequence (sequence);
	RETURN(status);
}


UINT2Sequence *XLALCreateUINT2Sequence (
	size_t length
)
{
	UINT2Sequence *new;
	UINT2 *data;

	new = LALMalloc(sizeof(*new));
	data = LALMalloc(length * sizeof(*data));
	if(!new || !data) {
		LALFree(new);
		LALFree(data);
		return(NULL);
	}

	new->data = data;
	new->length = length;

	return(new);
}


void LALCreateUINT2Sequence (
	LALStatus *status,
	UINT2Sequence **output,
	size_t length
)
{
	INITSTATUS(status, "LALCreateUINT2Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateUINT2Sequence (length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


UINT2Sequence *XLALCutUINT2Sequence (
	UINT2Sequence *sequence,
	size_t first,
	size_t length
)
{
	UINT2Sequence *new = NULL;

	if(sequence && sequence->data) {
		new = XLALCreateUINT2Sequence (length);
		if(new)
			memcpy(new->data, sequence->data + first, length * sizeof(*new->data));
	}

	return(new);
}


void LALCutUINT2Sequence (
	LALStatus *status,
	UINT2Sequence **output,
	UINT2Sequence *input,
	size_t first,
	size_t length
)
{
	INITSTATUS(status, "LALCutUINT2Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= input->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	*output = XLALCutUINT2Sequence (input, first, length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


UINT2Sequence *XLALCopyUINT2Sequence (
	UINT2Sequence *sequence
)
{
	return(XLALCutUINT2Sequence (sequence, 0, sequence->length));
}


void LALCopyUINT2Sequence (
	LALStatus *status,
	UINT2Sequence **output,
	UINT2Sequence *input
)
{
	INITSTATUS(status, "LALCopyUINT2Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCopyUINT2Sequence (input);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


static void ShiftUINT2Sequence (
	UINT2Sequence *sequence,
	int count
)
{
	memshift(sequence->data, sequence->length * sizeof(*sequence->data), count * (int) sizeof(*sequence->data));
}


void XLALShiftUINT2Sequence (
	UINT2Sequence *sequence,
	int count
)
{
	if(!sequence || !sequence->data || !count)
		return;

	ShiftUINT2Sequence (sequence, count);
	if((size_t) labs(count) >= sequence->length)
		memset(sequence->data, 0, sequence->length * sizeof(*sequence->data));
	else if(count > 0)
		memset(sequence->data, 0, count * sizeof(*sequence->data));
	else
		memset(sequence->data + sequence->length + count, 0, -count * sizeof(*sequence->data));
}


void LALShiftUINT2Sequence (
	LALStatus *status,
	UINT2Sequence *sequence,
	int count
)
{
	INITSTATUS(status, "LALShiftUINT2Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	XLALShiftUINT2Sequence (sequence, count);
	RETURN(status);
}


/* FIXME: this function does not take care to move the least number of bytes
 * possible.  A performance gain would be realized by being more careful. */
size_t XLALResizeUINT2Sequence (
	UINT2Sequence *sequence,
	int first,
	size_t length
)
{
	if(!sequence || !sequence->data)
		return(0);

	if(length > sequence->length) {
		/* need to increase memory */
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));

		if(sequence->data)
			ShiftUINT2Sequence (sequence, -first);
		else
			length = 0;
	} else {
		/* do not need to increase memory */
		ShiftUINT2Sequence (sequence, -first);
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));
		if(!sequence->data)
			length = 0;
	}

	return(sequence->length = length);
}


void LALResizeUINT2Sequence (
	LALStatus *status,
	UINT2Sequence *sequence,
	int first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALResizeUINT2Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	new_length = XLALResizeUINT2Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}


size_t XLALShrinkUINT2Sequence (
	UINT2Sequence *sequence,
	size_t first,
	size_t length
)
{
	return(XLALResizeUINT2Sequence (sequence, first, length));
}


void LALShrinkUINT2Sequence (
	LALStatus *status,
	UINT2Sequence *sequence,
	size_t first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALShrinkUINT2Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= sequence->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	new_length = XLALShrinkUINT2Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}




void XLALDestroyINT4Sequence (
	INT4Sequence *sequence
)
{
	if(sequence)
		LALFree(sequence->data);
	LALFree(sequence);
}


void LALDestroyINT4Sequence (
	LALStatus *status,
	INT4Sequence *sequence
)
{
	INITSTATUS(status, "LALDestroyINT4Sequence", SEQUENCEC);
	XLALDestroyINT4Sequence (sequence);
	RETURN(status);
}


INT4Sequence *XLALCreateINT4Sequence (
	size_t length
)
{
	INT4Sequence *new;
	INT4 *data;

	new = LALMalloc(sizeof(*new));
	data = LALMalloc(length * sizeof(*data));
	if(!new || !data) {
		LALFree(new);
		LALFree(data);
		return(NULL);
	}

	new->data = data;
	new->length = length;

	return(new);
}


void LALCreateINT4Sequence (
	LALStatus *status,
	INT4Sequence **output,
	size_t length
)
{
	INITSTATUS(status, "LALCreateINT4Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateINT4Sequence (length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


INT4Sequence *XLALCutINT4Sequence (
	INT4Sequence *sequence,
	size_t first,
	size_t length
)
{
	INT4Sequence *new = NULL;

	if(sequence && sequence->data) {
		new = XLALCreateINT4Sequence (length);
		if(new)
			memcpy(new->data, sequence->data + first, length * sizeof(*new->data));
	}

	return(new);
}


void LALCutINT4Sequence (
	LALStatus *status,
	INT4Sequence **output,
	INT4Sequence *input,
	size_t first,
	size_t length
)
{
	INITSTATUS(status, "LALCutINT4Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= input->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	*output = XLALCutINT4Sequence (input, first, length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


INT4Sequence *XLALCopyINT4Sequence (
	INT4Sequence *sequence
)
{
	return(XLALCutINT4Sequence (sequence, 0, sequence->length));
}


void LALCopyINT4Sequence (
	LALStatus *status,
	INT4Sequence **output,
	INT4Sequence *input
)
{
	INITSTATUS(status, "LALCopyINT4Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCopyINT4Sequence (input);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


static void ShiftINT4Sequence (
	INT4Sequence *sequence,
	int count
)
{
	memshift(sequence->data, sequence->length * sizeof(*sequence->data), count * (int) sizeof(*sequence->data));
}


void XLALShiftINT4Sequence (
	INT4Sequence *sequence,
	int count
)
{
	if(!sequence || !sequence->data || !count)
		return;

	ShiftINT4Sequence (sequence, count);
	if((size_t) labs(count) >= sequence->length)
		memset(sequence->data, 0, sequence->length * sizeof(*sequence->data));
	else if(count > 0)
		memset(sequence->data, 0, count * sizeof(*sequence->data));
	else
		memset(sequence->data + sequence->length + count, 0, -count * sizeof(*sequence->data));
}


void LALShiftINT4Sequence (
	LALStatus *status,
	INT4Sequence *sequence,
	int count
)
{
	INITSTATUS(status, "LALShiftINT4Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	XLALShiftINT4Sequence (sequence, count);
	RETURN(status);
}


/* FIXME: this function does not take care to move the least number of bytes
 * possible.  A performance gain would be realized by being more careful. */
size_t XLALResizeINT4Sequence (
	INT4Sequence *sequence,
	int first,
	size_t length
)
{
	if(!sequence || !sequence->data)
		return(0);

	if(length > sequence->length) {
		/* need to increase memory */
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));

		if(sequence->data)
			ShiftINT4Sequence (sequence, -first);
		else
			length = 0;
	} else {
		/* do not need to increase memory */
		ShiftINT4Sequence (sequence, -first);
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));
		if(!sequence->data)
			length = 0;
	}

	return(sequence->length = length);
}


void LALResizeINT4Sequence (
	LALStatus *status,
	INT4Sequence *sequence,
	int first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALResizeINT4Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	new_length = XLALResizeINT4Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}


size_t XLALShrinkINT4Sequence (
	INT4Sequence *sequence,
	size_t first,
	size_t length
)
{
	return(XLALResizeINT4Sequence (sequence, first, length));
}


void LALShrinkINT4Sequence (
	LALStatus *status,
	INT4Sequence *sequence,
	size_t first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALShrinkINT4Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= sequence->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	new_length = XLALShrinkINT4Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}




void XLALDestroyUINT4Sequence (
	UINT4Sequence *sequence
)
{
	if(sequence)
		LALFree(sequence->data);
	LALFree(sequence);
}


void LALDestroyUINT4Sequence (
	LALStatus *status,
	UINT4Sequence *sequence
)
{
	INITSTATUS(status, "LALDestroyUINT4Sequence", SEQUENCEC);
	XLALDestroyUINT4Sequence (sequence);
	RETURN(status);
}


UINT4Sequence *XLALCreateUINT4Sequence (
	size_t length
)
{
	UINT4Sequence *new;
	UINT4 *data;

	new = LALMalloc(sizeof(*new));
	data = LALMalloc(length * sizeof(*data));
	if(!new || !data) {
		LALFree(new);
		LALFree(data);
		return(NULL);
	}

	new->data = data;
	new->length = length;

	return(new);
}


void LALCreateUINT4Sequence (
	LALStatus *status,
	UINT4Sequence **output,
	size_t length
)
{
	INITSTATUS(status, "LALCreateUINT4Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateUINT4Sequence (length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


UINT4Sequence *XLALCutUINT4Sequence (
	UINT4Sequence *sequence,
	size_t first,
	size_t length
)
{
	UINT4Sequence *new = NULL;

	if(sequence && sequence->data) {
		new = XLALCreateUINT4Sequence (length);
		if(new)
			memcpy(new->data, sequence->data + first, length * sizeof(*new->data));
	}

	return(new);
}


void LALCutUINT4Sequence (
	LALStatus *status,
	UINT4Sequence **output,
	UINT4Sequence *input,
	size_t first,
	size_t length
)
{
	INITSTATUS(status, "LALCutUINT4Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= input->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	*output = XLALCutUINT4Sequence (input, first, length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


UINT4Sequence *XLALCopyUINT4Sequence (
	UINT4Sequence *sequence
)
{
	return(XLALCutUINT4Sequence (sequence, 0, sequence->length));
}


void LALCopyUINT4Sequence (
	LALStatus *status,
	UINT4Sequence **output,
	UINT4Sequence *input
)
{
	INITSTATUS(status, "LALCopyUINT4Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCopyUINT4Sequence (input);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


static void ShiftUINT4Sequence (
	UINT4Sequence *sequence,
	int count
)
{
	memshift(sequence->data, sequence->length * sizeof(*sequence->data), count * (int) sizeof(*sequence->data));
}


void XLALShiftUINT4Sequence (
	UINT4Sequence *sequence,
	int count
)
{
	if(!sequence || !sequence->data || !count)
		return;

	ShiftUINT4Sequence (sequence, count);
	if((size_t) labs(count) >= sequence->length)
		memset(sequence->data, 0, sequence->length * sizeof(*sequence->data));
	else if(count > 0)
		memset(sequence->data, 0, count * sizeof(*sequence->data));
	else
		memset(sequence->data + sequence->length + count, 0, -count * sizeof(*sequence->data));
}


void LALShiftUINT4Sequence (
	LALStatus *status,
	UINT4Sequence *sequence,
	int count
)
{
	INITSTATUS(status, "LALShiftUINT4Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	XLALShiftUINT4Sequence (sequence, count);
	RETURN(status);
}


/* FIXME: this function does not take care to move the least number of bytes
 * possible.  A performance gain would be realized by being more careful. */
size_t XLALResizeUINT4Sequence (
	UINT4Sequence *sequence,
	int first,
	size_t length
)
{
	if(!sequence || !sequence->data)
		return(0);

	if(length > sequence->length) {
		/* need to increase memory */
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));

		if(sequence->data)
			ShiftUINT4Sequence (sequence, -first);
		else
			length = 0;
	} else {
		/* do not need to increase memory */
		ShiftUINT4Sequence (sequence, -first);
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));
		if(!sequence->data)
			length = 0;
	}

	return(sequence->length = length);
}


void LALResizeUINT4Sequence (
	LALStatus *status,
	UINT4Sequence *sequence,
	int first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALResizeUINT4Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	new_length = XLALResizeUINT4Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}


size_t XLALShrinkUINT4Sequence (
	UINT4Sequence *sequence,
	size_t first,
	size_t length
)
{
	return(XLALResizeUINT4Sequence (sequence, first, length));
}


void LALShrinkUINT4Sequence (
	LALStatus *status,
	UINT4Sequence *sequence,
	size_t first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALShrinkUINT4Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= sequence->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	new_length = XLALShrinkUINT4Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}




void XLALDestroyINT8Sequence (
	INT8Sequence *sequence
)
{
	if(sequence)
		LALFree(sequence->data);
	LALFree(sequence);
}


void LALDestroyINT8Sequence (
	LALStatus *status,
	INT8Sequence *sequence
)
{
	INITSTATUS(status, "LALDestroyINT8Sequence", SEQUENCEC);
	XLALDestroyINT8Sequence (sequence);
	RETURN(status);
}


INT8Sequence *XLALCreateINT8Sequence (
	size_t length
)
{
	INT8Sequence *new;
	INT8 *data;

	new = LALMalloc(sizeof(*new));
	data = LALMalloc(length * sizeof(*data));
	if(!new || !data) {
		LALFree(new);
		LALFree(data);
		return(NULL);
	}

	new->data = data;
	new->length = length;

	return(new);
}


void LALCreateINT8Sequence (
	LALStatus *status,
	INT8Sequence **output,
	size_t length
)
{
	INITSTATUS(status, "LALCreateINT8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateINT8Sequence (length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


INT8Sequence *XLALCutINT8Sequence (
	INT8Sequence *sequence,
	size_t first,
	size_t length
)
{
	INT8Sequence *new = NULL;

	if(sequence && sequence->data) {
		new = XLALCreateINT8Sequence (length);
		if(new)
			memcpy(new->data, sequence->data + first, length * sizeof(*new->data));
	}

	return(new);
}


void LALCutINT8Sequence (
	LALStatus *status,
	INT8Sequence **output,
	INT8Sequence *input,
	size_t first,
	size_t length
)
{
	INITSTATUS(status, "LALCutINT8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= input->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	*output = XLALCutINT8Sequence (input, first, length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


INT8Sequence *XLALCopyINT8Sequence (
	INT8Sequence *sequence
)
{
	return(XLALCutINT8Sequence (sequence, 0, sequence->length));
}


void LALCopyINT8Sequence (
	LALStatus *status,
	INT8Sequence **output,
	INT8Sequence *input
)
{
	INITSTATUS(status, "LALCopyINT8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCopyINT8Sequence (input);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


static void ShiftINT8Sequence (
	INT8Sequence *sequence,
	int count
)
{
	memshift(sequence->data, sequence->length * sizeof(*sequence->data), count * (int) sizeof(*sequence->data));
}


void XLALShiftINT8Sequence (
	INT8Sequence *sequence,
	int count
)
{
	if(!sequence || !sequence->data || !count)
		return;

	ShiftINT8Sequence (sequence, count);
	if((size_t) labs(count) >= sequence->length)
		memset(sequence->data, 0, sequence->length * sizeof(*sequence->data));
	else if(count > 0)
		memset(sequence->data, 0, count * sizeof(*sequence->data));
	else
		memset(sequence->data + sequence->length + count, 0, -count * sizeof(*sequence->data));
}


void LALShiftINT8Sequence (
	LALStatus *status,
	INT8Sequence *sequence,
	int count
)
{
	INITSTATUS(status, "LALShiftINT8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	XLALShiftINT8Sequence (sequence, count);
	RETURN(status);
}


/* FIXME: this function does not take care to move the least number of bytes
 * possible.  A performance gain would be realized by being more careful. */
size_t XLALResizeINT8Sequence (
	INT8Sequence *sequence,
	int first,
	size_t length
)
{
	if(!sequence || !sequence->data)
		return(0);

	if(length > sequence->length) {
		/* need to increase memory */
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));

		if(sequence->data)
			ShiftINT8Sequence (sequence, -first);
		else
			length = 0;
	} else {
		/* do not need to increase memory */
		ShiftINT8Sequence (sequence, -first);
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));
		if(!sequence->data)
			length = 0;
	}

	return(sequence->length = length);
}


void LALResizeINT8Sequence (
	LALStatus *status,
	INT8Sequence *sequence,
	int first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALResizeINT8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	new_length = XLALResizeINT8Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}


size_t XLALShrinkINT8Sequence (
	INT8Sequence *sequence,
	size_t first,
	size_t length
)
{
	return(XLALResizeINT8Sequence (sequence, first, length));
}


void LALShrinkINT8Sequence (
	LALStatus *status,
	INT8Sequence *sequence,
	size_t first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALShrinkINT8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= sequence->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	new_length = XLALShrinkINT8Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}




void XLALDestroyUINT8Sequence (
	UINT8Sequence *sequence
)
{
	if(sequence)
		LALFree(sequence->data);
	LALFree(sequence);
}


void LALDestroyUINT8Sequence (
	LALStatus *status,
	UINT8Sequence *sequence
)
{
	INITSTATUS(status, "LALDestroyUINT8Sequence", SEQUENCEC);
	XLALDestroyUINT8Sequence (sequence);
	RETURN(status);
}


UINT8Sequence *XLALCreateUINT8Sequence (
	size_t length
)
{
	UINT8Sequence *new;
	UINT8 *data;

	new = LALMalloc(sizeof(*new));
	data = LALMalloc(length * sizeof(*data));
	if(!new || !data) {
		LALFree(new);
		LALFree(data);
		return(NULL);
	}

	new->data = data;
	new->length = length;

	return(new);
}


void LALCreateUINT8Sequence (
	LALStatus *status,
	UINT8Sequence **output,
	size_t length
)
{
	INITSTATUS(status, "LALCreateUINT8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateUINT8Sequence (length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


UINT8Sequence *XLALCutUINT8Sequence (
	UINT8Sequence *sequence,
	size_t first,
	size_t length
)
{
	UINT8Sequence *new = NULL;

	if(sequence && sequence->data) {
		new = XLALCreateUINT8Sequence (length);
		if(new)
			memcpy(new->data, sequence->data + first, length * sizeof(*new->data));
	}

	return(new);
}


void LALCutUINT8Sequence (
	LALStatus *status,
	UINT8Sequence **output,
	UINT8Sequence *input,
	size_t first,
	size_t length
)
{
	INITSTATUS(status, "LALCutUINT8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= input->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	*output = XLALCutUINT8Sequence (input, first, length);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


UINT8Sequence *XLALCopyUINT8Sequence (
	UINT8Sequence *sequence
)
{
	return(XLALCutUINT8Sequence (sequence, 0, sequence->length));
}


void LALCopyUINT8Sequence (
	LALStatus *status,
	UINT8Sequence **output,
	UINT8Sequence *input
)
{
	INITSTATUS(status, "LALCopyUINT8Sequence", SEQUENCEC);
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(input != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCopyUINT8Sequence (input);
	ASSERT(*output != NULL, status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	RETURN(status);
}


static void ShiftUINT8Sequence (
	UINT8Sequence *sequence,
	int count
)
{
	memshift(sequence->data, sequence->length * sizeof(*sequence->data), count * (int) sizeof(*sequence->data));
}


void XLALShiftUINT8Sequence (
	UINT8Sequence *sequence,
	int count
)
{
	if(!sequence || !sequence->data || !count)
		return;

	ShiftUINT8Sequence (sequence, count);
	if((size_t) labs(count) >= sequence->length)
		memset(sequence->data, 0, sequence->length * sizeof(*sequence->data));
	else if(count > 0)
		memset(sequence->data, 0, count * sizeof(*sequence->data));
	else
		memset(sequence->data + sequence->length + count, 0, -count * sizeof(*sequence->data));
}


void LALShiftUINT8Sequence (
	LALStatus *status,
	UINT8Sequence *sequence,
	int count
)
{
	INITSTATUS(status, "LALShiftUINT8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	XLALShiftUINT8Sequence (sequence, count);
	RETURN(status);
}


/* FIXME: this function does not take care to move the least number of bytes
 * possible.  A performance gain would be realized by being more careful. */
size_t XLALResizeUINT8Sequence (
	UINT8Sequence *sequence,
	int first,
	size_t length
)
{
	if(!sequence || !sequence->data)
		return(0);

	if(length > sequence->length) {
		/* need to increase memory */
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));

		if(sequence->data)
			ShiftUINT8Sequence (sequence, -first);
		else
			length = 0;
	} else {
		/* do not need to increase memory */
		ShiftUINT8Sequence (sequence, -first);
		sequence->data = LALRealloc(sequence->data, length * sizeof(*sequence->data));
		if(!sequence->data)
			length = 0;
	}

	return(sequence->length = length);
}


void LALResizeUINT8Sequence (
	LALStatus *status,
	UINT8Sequence *sequence,
	int first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALResizeUINT8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	new_length = XLALResizeUINT8Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}


size_t XLALShrinkUINT8Sequence (
	UINT8Sequence *sequence,
	size_t first,
	size_t length
)
{
	return(XLALResizeUINT8Sequence (sequence, first, length));
}


void LALShrinkUINT8Sequence (
	LALStatus *status,
	UINT8Sequence *sequence,
	size_t first,
	size_t length
)
{
	size_t new_length;

	INITSTATUS(status, "LALShrinkUINT8Sequence", SEQUENCEC);
	ASSERT(sequence != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	ASSERT(first + length <= sequence->length, status, LAL_RANGE_ERR, LAL_RANGE_MSG);
	new_length = XLALShrinkUINT8Sequence (sequence, first, length);
	ASSERT(new_length == length, status, LAL_FAIL_ERR, LAL_FAIL_MSG);
	RETURN(status);
}

